require("defines")

game.oninit(function()
	onLoad()
end)

game.onload(function()
	onLoad()
end)

game.onevent(defines.events.onbuiltentity, function(event)
	slipstreamBuilt(event.createdentity)
end)

game.onevent(defines.events.onrobotbuiltentity, function(event)
	slipstreamBuilt(event.createdentity)
end)

function onLoad()
	if not glob.slipstream then
	glob.slipstream = {}
	end
end

function slipstreamBuilt(entity)
	local direction = entity.name:match("^slipstream%-chest%-(%a+)$") -- Checks to see if the built entity is a slipstream chest and discovers its direction at the same time.
	
	if direction then -- Setting up the table of values.
		local new_chest = {
			entity = entity,
			inventory = entity.getinventory(defines.inventory.chest),
			direction = defines.direction[direction],
			collection = configureChestCollection(entity.position),
			spawn = configureChestSpawn(entity.position,direction)
		}
		new_chest.spawncatchment = {configureChestSpawnCatchment(new_chest.spawn,1), configureChestSpawnCatchment(new_chest.spawn,2)}
		
		--[[	Test block to report various bits of data.
		local first = tostring(new_chest["entity"])
		local second = tostring(new_chest.inventory)
		local third = tostring(new_chest["direction"])
		local fourth = "Coordindates: x:" .. new_chest.collection[1][1] .. " y:" .. new_chest.collection[1][2] .. ", x:" .. new_chest.collection[2][1] .. " y:" .. new_chest.collection[2][2]
		local fifth = "Spawn Coords: x:" .. new_chest.spawn[1][1] .. " y:" .. new_chest.spawn[1][2] .. ", x:" .. new_chest.spawn[2][1] .. "y: " .. new_chest.spawn[2][2]
		local sixth = "Spawn Catch 1 Coords: x:" .. new_chest.spawncatchment[1][1][1] .. " y:" .. new_chest.spawncatchment[1][1][2] .. ", x:" .. new_chest.spawncatchment[1][2][1] .. " y:" .. new_chest.spawncatchment[1][2][2]
		local seventh = "Spawn Catch 2 Coords: x:" .. new_chest.spawncatchment[2][1][1] .. " y:" .. new_chest.spawncatchment[2][1][2] .. ", x:" .. new_chest.spawncatchment[2][2][1] .. " y:" .. new_chest.spawncatchment[2][2][2]
		
		debugLog("Name: " .. new_chest.entity.name)
		debugLog("Entity: " .. first)
		debugLog("Inventory: " .. second)
		debugLog("Direction: " .. third)
		debugLog(fourth)
		debugLog(fifth)
		debugLog(sixth)
		debugLog(seventh)
		--]]
		table.insert(glob.slipstream, new_chest)
	end
	
end

function configureChestCollection(position) -- Configures the areas that the chest will search for new items. Default values should be 0.5
	return {
		{position.x - 0.5,position.y - 0.5},
		{position.x + 0.5,position.y + 0.5}
	}
end

function configureChestSpawn(position, direction) -- Configures the two locations that the chest will attempt to place items down on, depending on the chest's orientation.
	local spawnFunction = {
		north = function (position) return {{position.x - 0.25, position.y - 0.64},{position.x + 0.25, position.y - 0.64}} end,
		east = function (position) return {{position.x + 0.64, position.y - 0.25},{position.x + 0.64, position.y + 0.25}} end,
		south = function (position) return {{position.x + 0.25, position.y + 0.64},{position.x - 0.25, position.y + 0.64}} end,
		west = function (position) return {{position.x - 0.64, position.y + 0.25},{position.x - 0.64, position.y - 0.25}} end
	}
	
	return spawnFunction[direction](position)
end

function configureChestSpawnCatchment(position,number) -- Creates a pair of small search squares around where output items are deposited. These are searched to determine if anything is in them, since canplaceentity seems bugged.
	return {{position[number][1] - 0.2, position[number][2] - 0.2},{position[number][1] + 0.2, position[number][2] + 0.2}}
end

game.onevent(defines.events.ontick, function(event)
	if game.tick % 5 == 0 then
		for i,v in pairs(glob.slipstream) do
			if v.entity.valid then -- Checks whether the chest entity still exists, and if not, remove from the slipstream table. Otherwise, proceed.
				chestCollection(v)
				chestSpawn(v)
			else
				table.remove(glob.slipstream,i)
			end
		end
	end
end)

function chestCollection(storage)
	local pickup = game.findentitiesfiltered{area = storage.collection, name = "item-on-ground"}
	
	for _,v in pairs(pickup) do
		if storage.inventory.caninsert(v.stack) then
			storage.inventory.insert(v.stack)
			v.destroy()
		end
	end
end

function chestSpawn(storage)
	if not storage.inventory.isempty() then
		for i,v in pairs(storage.spawn) do
			local itemname = "foo"
			do	-- Compiles a list of names of items sitting in the chest.
				local keyset = {}
				local n=0
				
				for k,_ in pairs(storage.inventory.getcontents()) do
					n=n+1
					keyset[n]=k
				end
				
				itemname = keyset[1]
			end
			
			spawnposentity = game.findentitiesfiltered{area = {storage.spawncatchment[i][1], storage.spawncatchment[i][2]}, name="item-on-ground"}
			
			if next(spawnposentity) == nil then -- Tests whether anything is in the area where items are deposited by the chest, if not, proceed to putting stuff down.
				game.createentity{name="item-on-ground", position = v, stack = {name = itemname, count = 1}}
				storage.inventory.remove{name = itemname, count = 1}
			end
		end
	end
end

function debugLog(message)
	if false then -- set for debug
		game.player.print(message)
	end
end

function table.val_to_str ( v ) -- These aren't relevant to the mod except as debugging tools.
  if "string" == type( v ) then
    v = string.gsub( v, "\n", "\\n" )
    if string.match( string.gsub(v,"[^'\"]",""), '^"+$' ) then
      return "'" .. v .. "'"
    end
    return '"' .. string.gsub(v,'"', '\\"' ) .. '"'
  else
    return "table" == type( v ) and table.tostring( v ) or
      tostring( v )
  end
end

function table.key_to_str ( k )
  if "string" == type( k ) and string.match( k, "^[_%a][_%a%d]*$" ) then
    return k
  else
    return "[" .. table.val_to_str( k ) .. "]"
  end
end

function table.tostring( tbl )
  local result, done = {}, {}
  for k, v in ipairs( tbl ) do
    table.insert( result, table.val_to_str( v ) )
    done[ k ] = true
  end
  for k, v in pairs( tbl ) do
    if not done[ k ] then
      table.insert( result,
        table.key_to_str( k ) .. "=" .. table.val_to_str( v ) )
    end
  end
  return "{" .. table.concat( result, "," ) .. "}"
end